------------------------------------------------------------
-- Buffy by Sonaza
------------------------------------------------------------

local ADDON_NAME, SHARED = ...;

local _G = getfenv(0);

local LibStub = LibStub;
local A = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0");
_G[ADDON_NAME] = A;
SHARED[1] = A;

-- Text Strings
BUFFY_SET_BINDING_TEXT 		= "|cffeed028Set Buffy Cast Keybind|r";
BUFFY_CHOOSE_BINDING_TEXT	= "Choose a Binding";
BUFFY_PRESS_BUTTON_TEXT		= "Press a key to use it as the binding";
BUFFY_TEMPORARY_BIND_TEXT	= "Buffy will set a temporary binding when required and you can use any key you want without overwriting existing ones."
BUFFY_CANCEL_BINDING_TEXT 	= "Press Escape to Cancel";
BUFFY_ACCEPT_TEXT			= "Save";

local ALERT_TYPE_SPELL		= 0x1;
local ALERT_TYPE_ITEM		= 0x2;

-- Some buff category data
local BUFF_STATS 		= 0x001; -- RAID_BUFF_1 Stats
local BUFF_STAMINA 		= 0x002; -- RAID_BUFF_2 Stamina
local BUFF_ATTACK_POWER	= 0x004; -- RAID_BUFF_3 Attack Power
local BUFF_HASTE		= 0x008; -- RAID_BUFF_4 Haste
local BUFF_SPELL_POWER	= 0x010; -- RAID_BUFF_5 Spell Power
local BUFF_CRIT			= 0x020; -- RAID_BUFF_6 Critical Strike
local BUFF_MASTERY		= 0x040; -- RAID_BUFF_7 Mastery
local BUFF_MULTISTRIKE	= 0x080; -- RAID_BUFF_8 Multistrike
local BUFF_VERSATILITY	= 0x100; -- RAID_BUFF_9 Versatility

local BUFF_SPECIAL		= 0xFFF;

local BUFF_TYPES = {
	BUFF_STATS,
	BUFF_STAMINA,
	BUFF_ATTACK_POWER,
	BUFF_HASTE,
	BUFF_SPELL_POWER,
	BUFF_CRIT,
	BUFF_MASTERY,
	BUFF_MULTISTRIKE,
	BUFF_VERSATILITY,
};
A.BUFF_TYPES = BUFF_TYPES;

local BUFF_TYPE_NAMES = {
	[BUFF_STATS] 		= RAID_BUFF_1,
	[BUFF_STAMINA] 		= RAID_BUFF_2,
	[BUFF_ATTACK_POWER] = RAID_BUFF_3,
	[BUFF_HASTE] 		= RAID_BUFF_4,
	[BUFF_SPELL_POWER] 	= RAID_BUFF_5,
	[BUFF_CRIT] 		= RAID_BUFF_6,
	[BUFF_MASTERY] 		= RAID_BUFF_7,
	[BUFF_MULTISTRIKE] 	= RAID_BUFF_8,
	[BUFF_VERSATILITY] 	= RAID_BUFF_9,	
};
A.BUFF_TYPE_NAMES = BUFF_TYPE_NAMES;

local BUFF_SPELL_CATEGORIES = {
	[BUFF_STATS] 		= {},
	[BUFF_STAMINA] 		= {},
	[BUFF_ATTACK_POWER] = {},
	[BUFF_HASTE] 		= {},
	[BUFF_SPELL_POWER] 	= {},
	[BUFF_CRIT] 		= {},
	[BUFF_MASTERY] 		= {},
	[BUFF_MULTISTRIKE] 	= {},
	[BUFF_VERSATILITY] 	= {},
};
A.BUFF_SPELL_CATEGORIES = BUFF_SPELL_CATEGORIES;

local BUFFS = {};
local BUFF_CATEGORY_BY_SPELL = {};

local function AddBuffSpell(spell, categories, name)
	if(not spell or not categories) then return end
	
	if(name) then
		BUFFS[name] = spell;
	end
	
	BUFF_CATEGORY_BY_SPELL[spell] = categories;
	
	if(categories == BUFF_SPECIAL) then return end
	
	for _, category in pairs(BUFF_TYPES) do
		if(bit.band(categories, category) > 0) then
			tinsert(BUFF_SPELL_CATEGORIES[category], spell);
		end
	end
end

local function GetBuffCategory(spell)
	if(not spell) then return end
	
	return BUFF_CATEGORY_BY_SPELL[spell];
end

AddBuffSpell(20217,		BUFF_STATS, "BLESSING_OF_KINGS");
AddBuffSpell(115921,	BUFF_STATS, "LEGACY_OF_THE_EMPEROR");
AddBuffSpell(1126,		BUFF_STATS + BUFF_VERSATILITY, "MARK_OF_THE_WILD");
AddBuffSpell(116781,	BUFF_STATS + BUFF_CRIT, "LEGACY_OF_THE_WHITE_TIGER");
AddBuffSpell(160206,	BUFF_STATS, "LONEWOLF_PRIMATES");
AddBuffSpell(159988,	BUFF_STATS, "PET_DOG");
AddBuffSpell(160017,	BUFF_STATS, "PET_GORILLA");
AddBuffSpell(90363,		BUFF_STATS + BUFF_CRIT, "PET_SHALE_SPIDER");
AddBuffSpell(160077,	BUFF_STATS + BUFF_VERSATILITY, "PET_WORM");

AddBuffSpell(469, 		BUFF_STAMINA, "COMMANDING_SHOUT");
AddBuffSpell(21562,		BUFF_STAMINA, "POWER_WORD_FORTITUDE");
AddBuffSpell(166928,	BUFF_STAMINA, "BLOOD_PACT");
AddBuffSpell(160199,	BUFF_STAMINA, "LONEWOLF_BEAR");
AddBuffSpell(50256,		BUFF_STAMINA, "PET_BEAR");
AddBuffSpell(160014,	BUFF_STAMINA, "PET_GOAT");
AddBuffSpell(160003,	BUFF_STAMINA + BUFF_HASTE, "PET_RYLAK");
AddBuffSpell(90364,		BUFF_STAMINA + BUFF_SPELL_POWER, "PET_SILITHID");

AddBuffSpell(6673, 		BUFF_ATTACK_POWER, "BATTLE_SHOUT");
AddBuffSpell(57330, 	BUFF_ATTACK_POWER, "HORN_OF_WINTER");
AddBuffSpell(19506, 	BUFF_ATTACK_POWER, "TRUESHOT_AURA");

AddBuffSpell(55610,		BUFF_HASTE + BUFF_VERSATILITY, "UNHOLY_AURA");
AddBuffSpell(49868,		BUFF_HASTE + BUFF_MULTISTRIKE, "MIND_QUICKENING");
AddBuffSpell(116956,	BUFF_HASTE + BUFF_MASTERY, "GRACE_OF_AIR");
AddBuffSpell(113742,	BUFF_HASTE + BUFF_MULTISTRIKE, "SWIFTBLADES_CUNNING");
AddBuffSpell(160203,	BUFF_HASTE, "LONEWOLF_HYENA");
AddBuffSpell(128432,	BUFF_HASTE, "PET_HYENA");
AddBuffSpell(135678,	BUFF_HASTE, "PET_SPOREBAT");
AddBuffSpell(160074,	BUFF_HASTE, "PET_WASP");

AddBuffSpell(1459,		BUFF_SPELL_POWER + BUFF_CRIT, "ARCANE_BRILLIANCE");
AddBuffSpell(61316,		BUFF_SPELL_POWER + BUFF_CRIT, "DALARAN_BRILLIANCE");
AddBuffSpell(109773,	BUFF_SPELL_POWER + BUFF_MULTISTRIKE, "DARK_INTENT");
AddBuffSpell(160205,	BUFF_SPELL_POWER, "LONEWOLF_SERPENT");
AddBuffSpell(126309,	BUFF_SPELL_POWER, "PET_WATER_STRIDER");
AddBuffSpell(128433,	BUFF_SPELL_POWER, "PET_SERPENT");

AddBuffSpell(17007,		BUFF_CRIT, "LEADER_OF_THE_PACK");
AddBuffSpell(160200,	BUFF_CRIT, "LONEWOLF_RAPTOR");
AddBuffSpell(160052,	BUFF_CRIT, "PET_RAPTOR");
AddBuffSpell(24604,		BUFF_CRIT, "PET_WOLF");
AddBuffSpell(90309,		BUFF_CRIT, "PET_DEVILSAUR");
AddBuffSpell(126373,	BUFF_CRIT, "PET_QUILEN");

AddBuffSpell(155522, 	BUFF_MASTERY, "POWER_OF_THE_GRAVE");
AddBuffSpell(19740, 	BUFF_MASTERY, "BLESSING_OF_MIGHT");
AddBuffSpell(24907, 	BUFF_MASTERY, "MOONKIN_AURA");
AddBuffSpell(160198, 	BUFF_MASTERY, "LONEWOLF_CAT");
AddBuffSpell(93435, 	BUFF_MASTERY, "PET_CAT");
AddBuffSpell(160039, 	BUFF_MASTERY, "PET_HYDRA");
AddBuffSpell(160073, 	BUFF_MASTERY, "PET_TALLSTRIDER");
AddBuffSpell(128997, 	BUFF_MASTERY, "PET_SPIRIT_BEAST");

AddBuffSpell(166916,	BUFF_MULTISTRIKE, "WINDFLURRY");
AddBuffSpell(172968, 	BUFF_MULTISTRIKE, "LONEWOLF_DRAGONHAWK");
AddBuffSpell(50519, 	BUFF_MULTISTRIKE, "PET_BAT");
AddBuffSpell(34889, 	BUFF_MULTISTRIKE, "PET_DRAGONHAWK");
AddBuffSpell(24844, 	BUFF_MULTISTRIKE, "PET_WIND_SERPENT");
AddBuffSpell(58604, 	BUFF_MULTISTRIKE, "PET_CORE_HOUND");
AddBuffSpell(57386, 	BUFF_MULTISTRIKE + BUFF_VERSATILITY, "PET_CLEFTHOOF");

AddBuffSpell(167188,	BUFF_VERSATILITY, "INSPIRING_PRESENCE");
AddBuffSpell(167187,	BUFF_VERSATILITY, "SANCTITY_AURA");
AddBuffSpell(172967,	BUFF_VERSATILITY, "LONEWOLF_RAVAGER");
AddBuffSpell(50518,		BUFF_VERSATILITY, "PET_RAVAGER");
AddBuffSpell(159735,	BUFF_VERSATILITY, "PET_BIRD_OF_PREY");
AddBuffSpell(35290,		BUFF_VERSATILITY, "PET_BOAR");
AddBuffSpell(160045,	BUFF_VERSATILITY, "PET_PORCUPINE");
AddBuffSpell(173035,	BUFF_VERSATILITY, "PET_STAG");

AddBuffSpell(324, 		BUFF_SPECIAL, "SHAMAN_LIGHTNING_SHIELD");

AddBuffSpell(2823, 		BUFF_SPECIAL, "ROGUE_DEADLY_POISON");
AddBuffSpell(8679, 		BUFF_SPECIAL, "ROGUE_WOUND_POISON");
AddBuffSpell(3408, 		BUFF_SPECIAL, "ROGUE_CRIPPLING_POISON");
AddBuffSpell(108211,	BUFF_SPECIAL, "ROGUE_LEECHING_POISON");

AddBuffSpell(15473,		BUFF_SPECIAL, "PRIEST_SHADOWFORM");

AddBuffSpell(25780,		BUFF_SPECIAL, "PALADIN_RIGHTEOUS_FURY");

----------------------------------------------------------

local CONSUMABLE_FLASK	= 0x01;
local CONSUMABLE_RUNE	= 0x02;

local STAT_AGILITY 		= 0x001;
local STAT_STRENGTH 	= 0x002;
local STAT_INTELLECT	= 0x004;
local STAT_STAMINA		= 0x008;

local BUFF_ITEM_SPELLS = {};
local FLASKS = {};
local RUNES = {};

local function AddBuffItems(item_table, stat_type, items)
	if(not item_table or type(item_table) ~= "table") then return end
	if(not items or not stat_type) then return end
	
	item_table[stat_type] = items;
end

local function AddItemSpell(item, spell)
	if(not item or not spell) then return end
	
	BUFF_ITEM_SPELLS[item] = spell;
end

-- Greater Flasks first
AddBuffItems(FLASKS, STAT_AGILITY, 	{ 109153, 109145 });
AddItemSpell(109153, 156064); -- Greater Flask
AddItemSpell(109145, 156073);

AddBuffItems(FLASKS, STAT_STRENGTH,	{ 109156, 109148 });
AddItemSpell(109156, 156080); -- Greater Flask
AddItemSpell(109148, 156071);

AddBuffItems(FLASKS, STAT_INTELLECT,	{ 109155, 109147 });
AddItemSpell(109155, 156079); -- Greater Flask
AddItemSpell(109147, 156070);

AddBuffItems(FLASKS, STAT_STAMINA,	{ 109160, 109152 });
AddItemSpell(109160, 156084); -- Greater Flask
AddItemSpell(109152, 156077);

-- Non-consumable runes are listed first
AddBuffItems(RUNES, STAT_AGILITY, 	{ 128482, 128475, 118630 });
AddItemSpell(118630, 175456);

AddBuffItems(RUNES, STAT_STRENGTH,	{ 128482, 128475, 118631 });
AddItemSpell(118631, 175439);

AddBuffItems(RUNES, STAT_INTELLECT,	{ 128482, 128475, 118632 });
AddItemSpell(118632, 175457);

-- Reference table for main stats for classes and specs
local CLASS_STAT_PREFS = {
	["WARRIOR"]	= {
		[0] = { STAT_STRENGTH },
		[3] = function()
			if(A:PlayerHasTalent(7, 3) and GetShapeshiftForm() == 1) then
				return { STAT_STRENGTH };
			end
			
			return { STAT_STAMINA, STAT_STRENGTH };
		end,
	},
	["DEATHKNIGHT"]	= {
		[0] = { STAT_STRENGTH },
		[1] = { STAT_STAMINA, STAT_STRENGTH },
	},
	["PALADIN"]	= {
		[1] = { STAT_INTELLECT },
		[2] = { STAT_STAMINA, STAT_STRENGTH },
		[3] = { STAT_STRENGTH },
	},
	["MONK"] = {
		[1] = { STAT_STAMINA, STAT_AGILITY },
		[2] = { STAT_INTELLECT },
		[3] = { STAT_AGILITY },
	},
	["PRIEST"] = {
		[0] = { STAT_INTELLECT },
	},
	["SHAMAN"] = {
		[0] = { STAT_INTELLECT },
		[2] = { STAT_AGILITY },
	},
	["DRUID"] = {
		[1] = { STAT_INTELLECT },
		[2] = { STAT_AGILITY },
		[3] = { STAT_STAMINA, STAT_AGILITY },
		[4] = { STAT_INTELLECT },
	},
	["ROGUE"] = {
		[0] = { STAT_AGILITY },
	},
	["MAGE"] = {
		[0] = { STAT_INTELLECT },
	},
	["WARLOCK"] = {
		[0] = { STAT_INTELLECT },
	},
	["HUNTER"] = {
		[0] = { STAT_AGILITY },
	},
};

local function ltinsert(table, list)
	if(not table or type(table) ~= "table") then return false end
	if(not list or type(list) ~= "table") then return false end
	
	for key, value in ipairs(list) do
		-- tinsert(table, item)
		table[key] = value;
	end
end

function A:GetStatPreference(class, spec_index)
	if(not class or not spec_index) then return -1; end
	
	local statTypes = CLASS_STAT_PREFS[class][spec_index] or CLASS_STAT_PREFS[class][0];
	if(type(statTypes) == "function") then
		statTypes = statTypes();
	end
	
	local stats = {};
	ltinsert(stats, statTypes);
	
	if(#stats == 2 and stats[1] == STAT_STAMINA and self.db.global.ConsumablesRemind.SkipStamina) then
		-- Swap stats
		stats[1], stats[2] = stats[2], stats[1]
	end
	
	return stats;
end

--------------------------------------------

local function CanPoisonWeapons()
	-- return IsEquippedItemType("One-Hand");
	
	local mainhand = GetInventoryItemLink("player", 16);
	local mhSlot = mainhand and select(9, GetItemInfo(mainhand)) or "";
	
	local offhand = GetInventoryItemLink("player", 17);
	local ohSlot = offhand and select(9, GetItemInfo(offhand)) or "";
	
	return mhSlot == "INVTYPE_WEAPON" or mhSlot == "INVTYPE_WEAPONMAINHAND" or
		   ohSlot == "INVTYPE_WEAPON" or ohSlot == "INVTYPE_WEAPONOFFHAND";
end

local CLASS_CASTABLE_BUFFS = {
	["WARRIOR"]	= {
		[0]	= {
			{
				raidbuff = { BUFFS.BATTLE_SHOUT, BUFFS.COMMANDING_SHOUT, },
			},
		},
	},
	["DEATHKNIGHT"]	= {
		passive = {
			BUFFS.UNHOLY_AURA, BUFFS.POWER_OF_THE_GRAVE
		},
		[0]	= {
			{
				raidbuff = { BUFFS.HORN_OF_WINTER },
			},
		},
	},
	["PALADIN"]	= {
		passive = {
			BUFFS.SANCTITY_AURA,
		},
		[0]	= {
			{ 
				raidbuff = { BUFFS.BLESSING_OF_KINGS, BUFFS.BLESSING_OF_MIGHT },
			},
		},
		[2]	= {
			{ 
				raidbuff = { BUFFS.BLESSING_OF_KINGS, BUFFS.BLESSING_OF_MIGHT },
			},
			{
				selfbuff = { BUFFS.PALADIN_RIGHTEOUS_FURY, },
				condition = function()
					return A.db.global.Class.Paladin.RemindRighteousFury;
				end,
				description = "Missing Righteous Fury",
			},
		},
	},
	["MONK"] = {
		passive = {
			BUFFS.WINDFLURRY,
		},
		[1]	= {
			{
				raidbuff = { BUFFS.LEGACY_OF_THE_WHITE_TIGER, },
			},
		},
		[2]	= {
			{
				raidbuff = { BUFFS.LEGACY_OF_THE_EMPEROR, },
			},
		},
		[3]	= {
			{
				raidbuff = { BUFFS.LEGACY_OF_THE_WHITE_TIGER, },
			},
		},
	},
	["PRIEST"] = {
		[0]	= {
			{
				raidbuff = { BUFFS.POWER_WORD_FORTITUDE, },
			},
		},
		[3]	= {
			{
				raidbuff = { BUFFS.POWER_WORD_FORTITUDE, },
			},
			{
				selfbuff = function()
					return GetShapeshiftForm() ~= 1, BUFFS.PRIEST_SHADOWFORM;
				end,
				condition = function()
					return A.db.global.Class.Priest.RemindShadowform;
				end,
				description = "Not in Shadowform",
			},
		},
	},
	["SHAMAN"] = {
		passive = {
			BUFFS.GRACE_OF_AIR,	
		},
		[0]	= {
			{
				selfbuff = { BUFFS.SHAMAN_LIGHTNING_SHIELD, },
				condition = function()
					return A.db.global.Class.Shaman.RemindShield;
				end,
			},
		},
	},
	["DRUID"] = {
		passive = {
			BUFFS.LEADER_OF_THE_PACK,
		},
		[0]	= {
			{
				raidbuff = { BUFFS.MARK_OF_THE_WILD, },
			},
		},
	},
	["ROGUE"] = {
		passive = {
			BUFFS.SWIFTBLADES_CUNNING,
		},
		[0] = {
			{
				selfbuff = { BUFFS.ROGUE_DEADLY_POISON, BUFFS.ROGUE_WOUND_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableLethal and not A.db.global.Class.Rogue.WoundPoisonPriority;
				end,
				description = "Missing Lethal Poison",
			},
			{
				selfbuff = { BUFFS.ROGUE_WOUND_POISON, BUFFS.ROGUE_DEADLY_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableLethal and A.db.global.Class.Rogue.WoundPoisonPriority;
				end,
				description = "Missing Lethal Poison",
			},
			{
				noTalent = {3, 2},
				selfbuff = { BUFFS.ROGUE_CRIPPLING_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableNonlethal and not A.db.global.Class.Rogue.SkipCrippling;
				end,
				description = "Missing Non-Lethal Poison",
			},
			{
				hasTalent = {3, 2},
				selfbuff = { BUFFS.ROGUE_LEECHING_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableNonlethal;
				end,
				description = "Missing Non-Lethal Poison",
			},
		},
	},
	["MAGE"] = {
		[0]	= {
			{
				raidbuff = { BUFFS.DALARAN_BRILLIANCE, BUFFS.ARCANE_BRILLIANCE },
			},
		},
	},
	["WARLOCK"] = {
		passive = {
			BUFFS.BLOOD_PACT,
		},
		[0]	= {
			{
				raidbuff = { BUFFS.DARK_INTENT },
			},
		},
	},
	["HUNTER"] = {
		passive = {
			BUFFS.TRUESHOT_AURA,	
		},
		[0]	= {
			{
				hasTalent = {7, 3},
				raidbuff = {
					BUFFS.LONEWOLF_PRIMATES,
					BUFFS.LONEWOLF_DRAGONHAWK,
					BUFFS.LONEWOLF_RAPTOR,
					BUFFS.LONEWOLF_CAT,
					BUFFS.LONEWOLF_HYENA,
					BUFFS.LONEWOLF_RAVAGER,
					BUFFS.LONEWOLF_SERPENT,
					BUFFS.LONEWOLF_BEAR,
				},
				condition = function()
					return A.db.global.Class.Hunter.EnableLoneWolf;
				end,
			},
		},
		[1]	= {
			-- Beast Mastery doesn't have Lone Wolf
		},
	},
};

function A:OnInitialize()
	SLASH_BUFFY1	= "/bf";
	SLASH_BUFFY2	= "/buffy";
	SlashCmdList["BUFFY"] = function(command)
		A:OpenContextMenu(UIParent);
	end
	
	A:InitializeDatabase();
end

function A:OnEnable()
	A:RegisterEvent("UNIT_AURA");
	A:RegisterEvent("BAG_UPDATE_DELAYED");
	A:RegisterEvent("PLAYER_ALIVE");
	A:RegisterEvent("PLAYER_UNGHOST", "PLAYER_ALIVE");
	A:RegisterEvent("PLAYER_TALENT_UPDATE");
	A:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED");
	A:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
	A:RegisterEvent("PLAYER_REGEN_DISABLED");
	A:RegisterEvent("PLAYER_REGEN_ENABLED");
	A:RegisterEvent("ZONE_CHANGED");
	A:RegisterEvent("ZONE_CHANGED_NEW_AREA", "ZONE_CHANGED");
	A:RegisterEvent("COMPANION_UPDATE");
	
	A:InitializeDatabroker();

	A.IsFrameLocked = true;
	
	A:ScanTalents();
	A:UpdateBuffs();
	
	CreateFrame("Frame"):SetScript("OnUpdate", function(self, elapsed)
		self.elapsed = (self.elapsed or 0) + elapsed;
		
		if(self.elapsed >= 5.0 or (A.ShowingAlert and self.elapsed >= 1.0)) then
			A:UpdateBuffs();
			self.elapsed = 0;
		end
	end);
end

local GROUP_TYPE_SOLO 	= 1;
local GROUP_TYPE_PARTY 	= 2;
local GROUP_TYPE_RAID	= 3;

function A:GetGroupType()
	if(IsInRaid()) then
		return GROUP_TYPE_RAID;
	elseif(IsInGroup()) then
		return GROUP_TYPE_PARTY;
	end
	
	return GROUP_TYPE_SOLO;
end

function A:GetRealSpellID(spell_id)
	local spell_name = GetSpellInfo(spell_id);
	local name, _, _, _, _, _, realSpellID = GetSpellInfo(spell_name);
	
	return realSpellID or spell_id;
end

function A:ZONE_CHANGED()
	A:UpdateBuffs();
end

function A:UNIT_AURA()
	A:UpdateBuffs();
end

function A:COMPANION_UPDATE(event, companiontype)
	if(companiontype == "MOUNT") then
		A:UpdateBuffs();
	end
end

function A:PLAYER_SPECIALIZATION_CHANGED()
	A:PLAYER_TALENT_UPDATE();
end

function A:BAG_UPDATE_DELAYED()
	if(self.db.global.ConsumablesRemind.Enabled) then
		A:UpdateBuffs();
	end
end

function A:PLAYER_EQUIPMENT_CHANGED()
	A:UpdateBuffs();
end

function A:PLAYER_REGEN_DISABLED()
	A:UpdateBuffs();
	A.IsFrameLocked = true;
	BuffyFrame:EnableMouse(false);
	
	BuffySpellButtonFrame:SetAttribute("type1", nil);
end

function A:PLAYER_REGEN_ENABLED()
	A:UpdateBuffs();
	BuffyFrame:EnableMouse(true);
end

function A:PLAYER_TALENT_UPDATE()
	A:ScanTalents();
	A:UpdateBuffs();
end

function A:PLAYER_ALIVE()
	A:UpdateBuffs();
end

function A:ScanTalents()
	A.ActiveTalents = {};
	
	local numTiers = GetMaxTalentTier();
	local group = GetActiveSpecGroup();
	for tier = 1, numTiers do
		for column = 1, 3 do
			local talentID, _, _, selected = GetTalentInfo(tier, column, group);
			if(selected) then
				A.ActiveTalents[tier] = column;
				break;
			end
		end
	end	
end

function A:PlayerHasTalent(tier, column)
	return A.ActiveTalents and A.ActiveTalents[tier] and A.ActiveTalents[tier] == column;
end

-- Special abstraction hack to give fake info about player's hidden passive buffs as if they had them
function A:UnitAura(unit, spell_name, rank, flags)
	local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster = UnitAura(unit, spell_name, rank, flags);
	
	if(A:GetGroupType() == GROUP_TYPE_SOLO and unit == "player" and name == nil) then
		local _, PLAYER_CLASS = UnitClass("player");
		local passiveBuffs = CLASS_CASTABLE_BUFFS[PLAYER_CLASS].passive;
		
		if(passiveBuffs ~= nil) then
			for i, spell_id in pairs(passiveBuffs) do
				if(IsSpellKnown(spell_id)) then
					local name, rank, icon = GetSpellInfo(spell_id);
					if(name == spell_name) then
						return name, rank, icon, 0, nil, 0, 0, "player";
					end
				end
			end
		end
	end
	
	return name, rank, icon, count, debuffType, duration, expirationTime, unitCaster;
end

function A:UnitHasBuff(unit, spell)
	if(not unit or not spell) then return false end
	
	local realSpellID = A:GetRealSpellID(spell);
	local spell_name = GetSpellInfo(realSpellID);
	if(not spell_name) then return false end
	
	local name, _, _, _, _, duration, expirationTime, unitCaster = A:UnitAura(unit, spell_name, nil, "HELPFUL");
	if(not name) then
		return false;
	end
	
	return true, unitCaster == "player", unitCaster, expirationTime - GetTime();
end

function A:UnitInRange(unit)
	if(not unit) then return false end
	
	if(unit == "player") then return not UnitIsDeadOrGhost(unit) end
	
	if(not UnitIsVisible(unit) or not UnitIsConnected(unit) or UnitIsDeadOrGhost(unit)) then
		return false;
	end
	
	local unitInRange = nil;
	local distance, distanceChecked = UnitDistanceSquared(unit);
	if(distanceChecked) then
		unitInRange = distance <= 900;
	end
	
	return unitInRange;
end

local partyUnitID = { "player", "party1", "party2", "party3", "party4" };

function A:GetUnitID(group_type, index)
	if(group_type == GROUP_TYPE_SOLO or group_type == GROUP_TYPE_PARTY) then
		return partyUnitID[index];
	elseif(group_type == GROUP_TYPE_RAID) then
		return string.format("raid%d", index);
	end
	
	return nil;
end

local function GroupIterator()
	local index = 0;
	local groupType = A:GetGroupType();
	local numGroupMembers = GetNumGroupMembers();
	if(groupType == GROUP_TYPE_SOLO) then numGroupMembers = 1 end
	
	return function()
		index = index + 1;
		if(index <= numGroupMembers) then
			return index, A:GetUnitID(groupType, index);
		end
	end
end

function A:UnitIsPet(searchunit)
	if(not searchunit) then return false end
	
	local pets = {};
	for index, unit in GroupIterator() do
		if(UnitExists(unit .. "pet") and UnitIsUnit(searchunit, unit .. "pet")) then
			return true, UnitName(unit), UnitName(unit .. "pet");
		end
	end
	
	return false;
end

function A:ScanBuffs()
	local missingBuffs = {};
	
	for category, spells in pairs(BUFF_SPELL_CATEGORIES) do
		for index, unit in GroupIterator() do
			local buffFound = false;
			local isCastByPlayer;
			
			for _, spell in ipairs(spells) do
				local hasBuff, isInRange;
				hasBuff, isCastByPlayer = A:UnitHasBuff(unit, spell);
				
				if(hasBuff) then
					buffFound = true;
					break;
				end
			end
			
			if(not buffFound) then
				if(not missingBuffs[unit]) then
					missingBuffs[unit] = {};
				end
				
				tinsert(missingBuffs[unit], category);
			end
		end
	end
	
	return missingBuffs;
end

function A:ScanCurrentBuffs()
	local buffs = {};
	
	for category, spells in pairs(BUFF_SPELL_CATEGORIES) do
		local hasBuff, isCastByPlayer, buff_spell, caster = A:UnitHasCategoryBuff("player", category);
		
		if(not hasBuff) then
			buffs[category] = false;
		else
			local name, class;
			if(caster ~= nil) then
				name = UnitName(caster);
				_, class = UnitClass(caster);
			else
				name = "Unknown";
				class = "PRIEST";
			end
			
			buffs[category] = {
				buff = buff_spell,
				casterName = name,
				casterClass = class,
			};
		end
	end
	
	return buffs;
end

function A:ScanMissingBuff(spell)
	if(not spell) then return nil end
	
	local spellCategory = BUFF_CATEGORY_BY_SPELL[spell];
	if(not spellCategory) then return nil end
	
	local checkCategories = {};
	
	for _, category in pairs(BUFF_TYPES) do
		if(bit.band(spellCategory, category) > 0) then
			tinsert(checkCategories, category);
		end
	end
	
	local unitsMissingBuffs = {};
	local buffCastByPlayer;
	
	for _, category in ipairs(checkCategories) do
		for index, unit in GroupIterator() do
			local buffFound = false;
			local isCastByPlayer;
			
			for _, spell in ipairs(BUFF_SPELL_CATEGORIES[category]) do
				local hasBuff;
				hasBuff, isCastByPlayer = A:UnitHasBuff(unit, spell);
				
				if(hasBuff) then
					buffFound = true;
					buffCastByPlayer = isCastByPlayer;
					break;
				end
			end
			
			if(not buffFound and A:UnitInRange(unit)) then
				tinsert(unitsMissingBuffs, unit);
			end
		end
	end
	
	return unitsMissingBuffs, buffCastByPlayer;
end

-- function A:UnitHasCategoryBuff(unit, category)
-- 	if(not unit or not category) then return nil end
-- 	if(not A:UnitInRange(unit)) then return nil end
	
-- 	for _, spell in ipairs(BUFF_SPELL_CATEGORIES[category]) do
-- 		local hasBuff, isCastByPlayer, caster, remaining = A:UnitHasBuff(unit, spell);
-- 		if(hasBuff) then
-- 			return hasBuff, isCastByPlayer, spell, caster, A:WillBuffExpireSoon(remaining);
-- 		end
-- 	end
	
-- 	return false;
-- end

function A:UnitHasCategoryBuff(unit, category)
	if(not unit or not category) then return nil end
	if(not A:UnitInRange(unit)) then return nil end
	
	local hasBuff, isCastByPlayer, buffSpell, remaining;
	
	for _, spell in ipairs(BUFF_SPELL_CATEGORIES[category]) do
		hasBuff, isCastByPlayer, caster, remaining = A:UnitHasBuff(unit, spell);
		if(hasBuff) then
			buffSpell = spell;
			break
		end
	end
	
	return hasBuff, isCastByPlayer, buffSpell, caster, A:WillBuffExpireSoon(remaining or 0);
end

local BUFF_STATE_NO_BUFF 			= 0;
local BUFF_STATE_HAS_BUFF 			= 1;
local BUFF_STATE_HAS_BUFF_BY_PLAYER = 2;

function A:ScanBuffList(bufflist)
	if(not bufflist or type(bufflist) ~= "table") then return nil end
	
	local partyBuffState = {};
	
	-- print("A:ScanBuffList");
	for _, spell in ipairs(bufflist) do
		if(not spell) then return nil end
		
		-- print(_, spell);
		
		local spellCategory = BUFF_CATEGORY_BY_SPELL[spell];
		if(not spellCategory) then return nil end
		
		local checkCategories = {};
		
		for _, category in pairs(BUFF_TYPES) do
			if(bit.band(spellCategory, category) > 0) then
				tinsert(checkCategories, category);
			end
		end
		
		for _, category in ipairs(checkCategories) do
			-- print("Category", _, category);
			
			if(not partyBuffState[category]) then
				partyBuffState[category] = {};
			end
			
			for index, unit in GroupIterator() do
				local hasBuff, isCastByPlayer, buffSpell, caster, willExpireSoon = A:UnitHasCategoryBuff(unit, category);
				
				if(hasBuff ~= nil) then
					local buffState = BUFF_STATE_NO_BUFF;
					
					if(hasBuff) then
						if(not isCastByPlayer) then
							buffState = BUFF_STATE_HAS_BUFF;
						else
							buffState = BUFF_STATE_HAS_BUFF_BY_PLAYER;
						end
					end
					
					partyBuffState[category][unit] = {
						state = buffState,
						expiring = willExpireSoon,
					};
				end
			end
		end
	end
	
	return partyBuffState;
end

-- Buffy:GetPlayerRaidBuffByCategory(1)
function A:GetPlayerRaidBuffByCategory(search_category)
	local _, PLAYER_CLASS = UnitClass("player");
	local spec = GetSpecialization();
	
	-- Try to find a spec based list and if not found default to basic list
	local buffs = CLASS_CASTABLE_BUFFS[PLAYER_CLASS][spec] or CLASS_CASTABLE_BUFFS[PLAYER_CLASS][0];
	if(not buffs) then return false end
	
	for _, data in ipairs(buffs) do
		if(data.raidbuff) then
			for _, spell in ipairs(data.raidbuff) do
				local spellCategory = BUFF_CATEGORY_BY_SPELL[spell];
				if(not spellCategory) then return nil end
				
				if(bit.band(spellCategory, search_category) > 0 and IsSpellKnown(spell)) then
					return spell;
				end
			end
		end
	end
	
	return nil;
end 

function A:UnitHasSomeBuff(unit, bufflist)
	for _, buff in ipairs(bufflist) do
		if(IsSpellKnown(buff)) then
			local hasBuff, isCastByPlayer, caster, remaining = A:UnitHasBuff(unit, buff);
			if(hasBuff) then
				return true, buff, A:WillBuffExpireSoon(remaining);
			end
		end
	end
	
	return false, bufflist[1], false;
end

local function IsReallyMounted()
	-- IsMounted() or 
	return IsFlying() or UnitOnTaxi("player") or UnitInVehicle("player");
end

function A:WillBuffExpireSoon(remaining)
	return remaining > 0 and remaining <= self.db.global.ExpirationAlertThreshold;
end

function A:PlayerHasConsumableBuff(consumable_type, preferredStat)
	local hasBuff, consumableID, buffExpiring;
	
	local consumablesList = nil;
	
	if(consumable_type == CONSUMABLE_FLASK) then
		consumablesList = FLASKS;
	elseif(consumable_type == CONSUMABLE_RUNE) then
		consumablesList = RUNES;
	end
	
	if(consumablesList ~= nil and type(consumablesList) == "table") then
		for _, itemID in ipairs(consumablesList[preferredStat]) do
			local spellID = BUFF_ITEM_SPELLS[itemID];
			local hasBuff, _, _, remaining = A:UnitHasBuff("player", spellID);
			if(hasBuff) then
				return true, itemID, spellID, A:WillBuffExpireSoon(remaining);
			end
		end
	end
	
	return false, nil, nil, false;
end

local INSTANCETYPE_RAID 	= 0x1;
local INSTANCETYPE_DUNGEON 	= 0x2;

function A:PlayerInDraenorInstance(includeDungeons)
	includeDungeons = includeDungeons or false;
	
	local name, instanceType, difficultyID, difficultyName, maxPlayers, dynamicDifficulty, isDynamic, mapID, instanceGroupSize = GetInstanceInfo();
	
	if(instanceType ~= "raid" and instanceType ~= "party") then return false end
	
	local instanceMapIDs = {
		[1228] = INSTANCETYPE_RAID, -- Highmaul
		[1205] = INSTANCETYPE_RAID, -- Blackrock Foundry
		[1448] = INSTANCETYPE_RAID, -- Hellfire Citadel
		
		[1182] = INSTANCETYPE_DUNGEON, -- Auchindoun
		[1175] = INSTANCETYPE_DUNGEON, -- Bloodmaul Slag Mines
		[1208] = INSTANCETYPE_DUNGEON, -- Grimrail Depot
		[1195] = INSTANCETYPE_DUNGEON, -- Iron Docks
		[1176] = INSTANCETYPE_DUNGEON, -- Shadowmoon Burial Grounds
		[1209] = INSTANCETYPE_DUNGEON, -- Skyreach
		[1279] = INSTANCETYPE_DUNGEON, -- The Everbloom
		[1358] = INSTANCETYPE_DUNGEON, -- Upper Blackrock Spire
	};
	
	if(instanceMapIDs[mapID] ~= nil) then
		return instanceMapIDs[mapID] == INSTANCETYPE_RAID or (includeDungeons and instanceMapIDs[mapID] == INSTANCETYPE_DUNGEON);
	end
	
	return false;
end

function A:FindBestConsumableItem(consumable_type, preferredStat)
	local consumablesList = nil;
	
	if(consumable_type == CONSUMABLE_FLASK) then
		consumablesList = FLASKS;
	elseif(consumable_type == CONSUMABLE_RUNE) then
		consumablesList = RUNES;
	end
	
	if(consumablesList ~= nil and type(consumablesList) == "table") then
		for _, itemID in ipairs(consumablesList[preferredStat]) do
			local count = GetItemCount(itemID);
			if(count > 0) then
				return itemID, count;
			end
		end
	end
	
	return nil, 0;
end

function A:UpdateBuffs()
	A:UpdateDatabrokerText();
	
	local inInstance, instanceType = A:PlayerInInstance();
	
	local hideAlert = not self.db.global.Enabled or
					  (self.db.global.DisableOutside and not inInstance) or
					  (not self.db.global.ShowInCombat and InCombatLockdown()) or
					  (not self.db.global.ShowWhileMounted and IsReallyMounted());
	
	local canShowExpiration = not InCombatLockdown() or (InCombatLockdown() and not self.db.global.NoExpirationAlertInCombat);
	
	if(hideAlert) then
		A:HideBuffyAlert();
		return;
	end
	
	local _, PLAYER_CLASS = UnitClass("player");
	local PLAYER_SPEC = GetSpecialization();
	local PLAYER_LEVEL = UnitLevel("player");
	
	-- Try to find a spec based list and if not found default to basic list
	local buffs = CLASS_CASTABLE_BUFFS[PLAYER_CLASS][PLAYER_SPEC] or CLASS_CASTABLE_BUFFS[PLAYER_CLASS][0];
	if(not buffs) then return false end
	
	local playerCastExists = false;
	local shouldCastSpell = false;
	local castSpellID;
	
	local infoUnits = {};
	
	for _, data in ipairs(buffs) do
		local valid = true;
		
		if(data.hasTalent) then
			valid = valid and A:PlayerHasTalent(data.hasTalent[1], data.hasTalent[2]);
		end
		
		if(data.noTalent) then
			valid = valid and not A:PlayerHasTalent(data.noTalent[1], data.noTalent[2]);
		end
		
		if(data.condition and type(data.condition) == "function") then
			valid = valid and data.condition();
		end
		
		if(valid) then
			if(data.raidbuff) then
				local buffStates = A:ScanBuffList(data.raidbuff);
				
				local playerBuffCategory = 0;
				local playerBuffExists = false;
				local playerBuffExpiring = false;
				
				for category, unitBuffs in pairs(buffStates) do
					for unit, buffState in pairs(unitBuffs) do
						if(buffState.state == BUFF_STATE_HAS_BUFF_BY_PLAYER) then
							playerBuffExists = true;
							playerBuffCategory = category;
							
							playerBuffExpiring = buffState.expiring;
							break;
						end
					end
					
					if(playerBuffExists) then break end
				end
				
				local missingCategory = 0;
				local everybodyHasBuff = true;
				
				if(playerBuffExists) then
					for unit, buffState in pairs(buffStates[playerBuffCategory]) do
						if(buffState.state ~= BUFF_STATE_HAS_BUFF_BY_PLAYER) then
							everybodyHasBuff = false;
							missingCategory = playerBuffCategory;
							infoUnits[unit] = unit;
						end
					end
					
					if(not everybodyHasBuff) then
						for category, unitBuffs in pairs(buffStates) do
							if(category ~= missingCategory) then
								for unit, buffState in pairs(unitBuffs) do
									if(buffState.state == BUFF_STATE_HAS_BUFF_BY_PLAYER) then
										infoUnits[unit] = nil;
									end
								end
							end
						end
						
						local unitsMissing = 0;
						for _, _ in pairs(infoUnits) do
							unitsMissing = unitsMissing + 1;
						end
						
						if(unitsMissing == 0) then
							everybodyHasBuff = true;
						end
					end
				else
					for category, unitBuffs in pairs(buffStates) do
						for unit, buffState in pairs(unitBuffs) do
							if(buffState.state == BUFF_STATE_NO_BUFF) then
								everybodyHasBuff = false;
								missingCategory = category;
								infoUnits[unit] = unit;
							end
						end
						
						if(not everybodyHasBuff) then break end
					end
				end
				
				if(not everybodyHasBuff) then
					local buffSpell = A:GetPlayerRaidBuffByCategory(missingCategory);
					
					local units = {};
					for unit, _ in pairs(infoUnits) do
						tinsert(units, unit);
					end
					
					A:ShowBuffyAlert(ALERT_TYPE_SPELL, buffSpell, {
						units = units,
						category = missingCategory,
						expiring = playerBuffExpiring,
						remaining = A:GetBuffRemaining("player", buffSpell),
					});
					
					return;
				elseif(playerBuffExpiring and canShowExpiration) then
					local buffSpell = A:GetPlayerRaidBuffByCategory(playerBuffCategory);
					
					A:ShowBuffyAlert(ALERT_TYPE_SPELL, buffSpell, {
						category = playerBuffCategory,
						expiring = playerBuffExpiring,
						remaining = A:GetBuffRemaining("player", buffSpell),
					});
					return;
				end
			end
			
			if(data.selfbuff and not UnitIsDeadOrGhost("player")) then
				if(type(data.selfbuff) == "function") then
					local shouldAlert, buffSpell = data.selfbuff();
					
					if(shouldAlert and IsSpellKnown(buffSpell)) then
						A:ShowBuffyAlert(ALERT_TYPE_SPELL, buffSpell, {
							description = data.description,
						});
						return;
					end
				else
					local hasBuff, buffSpell, buffExpiring = A:UnitHasSomeBuff("player", data.selfbuff);
					
					if((not hasBuff or (buffExpiring and canShowExpiration)) and IsSpellKnown(buffSpell)) then
						A:ShowBuffyAlert(ALERT_TYPE_SPELL, buffSpell, {
							description = data.description,
							expiring = buffExpiring,
							remaining = A:GetBuffRemaining("player", buffSpell),
						});
						
						return;
					end
				end
			end
		end
	end
	
	-- Clear binds before flask remind since user may not have keybinds enabled
	A:ClearTempBind();
	
	--------------------------------------------------
	-- Consumables check
	
	local canShowConsumableAlert = not InCombatLockdown() or (InCombatLockdown() and not self.db.global.ConsumablesRemind.NotInCombat);
	
	if(self.db.global.ConsumablesRemind.Enabled and canShowConsumableAlert and PLAYER_LEVEL >= 90 and A:PlayerInDraenorInstance(self.db.global.ConsumablesRemind.EnableInParty)) then
		local preferredStats = A:GetStatPreference(PLAYER_CLASS, PLAYER_SPEC);
		
		if(self.db.global.ConsumablesRemind.Flasks) then
			-- Loop through preferred stat types
			for _, preferredStat in ipairs(preferredStats) do
				local playerHasFlask, itemID, spellID, buffExpiring = A:PlayerHasConsumableBuff(CONSUMABLE_FLASK, preferredStat);
				
				if(not playerHasFlask or (buffExpiring and canShowExpiration)) then
					-- Check if player has any flasks in their inventory
					local bestFlaskID, count = A:FindBestConsumableItem(CONSUMABLE_FLASK, preferredStat);
					
					if(bestFlaskID) then
						A:ShowBuffyAlert(ALERT_TYPE_ITEM, bestFlaskID, {
							spellID = spellID,
							expiring = buffExpiring,
							count = count,
							remaining = A:GetBuffRemaining("player", spellID),
						});
						return;
					end
				end
				
				if(playerHasFlask and not buffExpiring) then break end
			end
		end
		
		if(self.db.global.ConsumablesRemind.Runes and PLAYER_LEVEL >= 100) then
			-- Loop through preferred stat types
			for _, preferredStat in ipairs(preferredStats) do
				-- There are no stamina runes, skip!
				if(preferredStat ~= STAT_STAMINA) then
					local playerHasRune, itemID, spellID, buffExpiring = A:PlayerHasConsumableBuff(CONSUMABLE_RUNE, preferredStat);
					
					if(not playerHasRune or (buffExpiring and canShowExpiration)) then
						-- Check if player has any runes in their inventory
						local bestRuneID, count = A:FindBestConsumableItem(CONSUMABLE_RUNE, preferredStat);
						
						local isInfiniteRune = (bestRuneID == 128482 or bestRuneID == 128475);
						local cooldownExpired = true;
						
						-- Cooldown check for the non-consumable rune which has one
						if(isInfiniteRune) then
							local startTime, duration = GetItemCooldown(bestRuneID);
							cooldownExpired = (startTime == 0 and duration == 0);
							
							-- Reset count since infinite rune is not consumed
							count = nil;
						end
						
						if(bestRuneID and cooldownExpired) then
							A:ShowBuffyAlert(ALERT_TYPE_ITEM, bestRuneID, {
								spellID = spellID,
								expiring = buffExpiring,
								count = count,
								remaining = A:GetBuffRemaining("player", spellID),
							});
							return;
						end
					end
				
					if(playerHasRune and not buffExpiring) then break end
				end
			end
		end
	end
	
	A:HideBuffyAlert();
end

function A:GetBuffRemaining(unit, spellID)
	if(not unit or not spellID) then return 0 end
	
	local realSpellID = A:GetRealSpellID(spellID);
	local spellName = GetSpellInfo(realSpellID);
	
	local name, _, _, _, _, _, expirationTime = UnitAura(unit, spellName, nil, "HELPFUL");
	if(name) then
		return expirationTime - GetTime();
	end
	
	return 0;
end

function A:FormatTime(t, literal)
	local seconds = t % 60;
	local minutes = math.floor(t / 60);
	
	local timestr = "";
	
	local M  = "%d|cfff1da54m|r";
	local  S = "%d|cfff1da54s|r";
	local MS = string.format("%s %s", M, S);
	
	if(literal) then
		M  = "%d minute" .. (minutes ~= 1 and "s" or "");
		 S = "%d second" .. (seconds ~= 1 and "s" or "");
		
		MS = string.format("%s %s", M, S);
	end
	
	if(minutes > 0 and seconds == 0 and literal) then
		return string.format(M, minutes);
	elseif(minutes > 0) then
		return string.format(MS, minutes, seconds);
	else
		return string.format(S, seconds);
	end
end

function A:ShowBuffyAlert(alert_type, id, data)
	if(not id) then return end
	
	A.ShowingAlert = true;
	
	local _;
	local name, icon;
	
	if(alert_type == ALERT_TYPE_SPELL) then
		local realSpellID = A:GetRealSpellID(id);
		name, _, icon = GetSpellInfo(realSpellID);
		BuffyFrame.icon.texture:SetTexture(icon);
		
		if(not InCombatLockdown()) then
			BuffySpellButtonFrame:SetAttribute("type1", "spell");
			BuffySpellButtonFrame:SetAttribute("spell1", name);
			BuffySpellButtonFrame:SetAttribute("unit", "player");
			A:SetTempBind();
		end
		
		BuffyFrame.buffname:SetFormattedText("|cfff1da54Cast|r|n%s", name or "<Error>");
	elseif(alert_type == ALERT_TYPE_ITEM) then
		name, _, _, _, _, _, _, _, _, icon = GetItemInfo(id);
		BuffyFrame.icon.texture:SetTexture(icon);
		
		if(not InCombatLockdown() and self.db.global.ConsumablesRemind.KeybindEnabled) then
			BuffySpellButtonFrame:SetAttribute("type1", "item");
			BuffySpellButtonFrame:SetAttribute("item1", name);
			BuffySpellButtonFrame:SetAttribute("unit", "player");
			A:SetTempBind();
		end
		
		BuffyFrame.buffname:SetFormattedText("|cfff1da54Use|r|n%s", name or "<Error>");
	end
	
	BuffyFrame.description:Hide();
	
	if(data) then
		if(not data.expiring and data.units and A:GetGroupType() ~= GROUP_TYPE_SOLO and data.category) then
			local numUnits = #data.units;
			BuffyFrame.description:SetFormattedText("%d player%s missing %s", numUnits, numUnits == 1 and "" or "s", BUFF_TYPE_NAMES[data.category]);
			BuffyFrame.description:Show();
		elseif(data.expiring and data.remaining) then
			if(alert_type == ALERT_TYPE_ITEM and data.count) then
				BuffyFrame.description:SetFormattedText("Expiring in %s / You have |cfff1da54%d|r", A:FormatTime(data.remaining), data.count);
			else
				BuffyFrame.description:SetFormattedText("Expiring in %s", A:FormatTime(data.remaining));
			end
			BuffyFrame.description:Show();
		elseif(alert_type == ALERT_TYPE_ITEM and data.count) then
			BuffyFrame.description:SetFormattedText("You have |cfff1da54%d|r in your inventory", data.count);
			BuffyFrame.description:Show();
		elseif(data.description) then
			BuffyFrame.description:SetText(data.description);
			BuffyFrame.description:Show();
		end
	end
	
	if(not BuffyFrame:IsShown()) then
		BuffyFrame.fadein:Play();
	end
end

function A:HideBuffyAlert()
	A.ShowingAlert = false;
	
	if(not InCombatLockdown()) then
		BuffySpellButtonFrame:SetAttribute("type1", nil);
		A:ClearTempBind();
	end
	
	if(BuffyFrame:IsShown()) then
		BuffyFrame.fadeout:Play();
	end
end

function A:PlayerInInstance()
	local name, instanceType = GetInstanceInfo();
	
	if(instanceType == "none" or C_Garrison.IsOnGarrisonMap()) then
		return false;
	end
	
	return true, instanceType;
end

function BuffyFrame_OnShow(self)
	local scale = UIParent:GetEffectiveScale();
	self:SetScale(1.0 / scale * 0.92);
	
	BuffyFrame_RestorePosition();
	
	-- self:ClearAllPoints();
	-- self:SetPoint("LEFT", UIParent, "CENTER", -22, GetScreenHeight() * 0.18);
end


function BuffyFrame_OnMouseDown()
	if(A.IsFrameLocked) then return end
	
	CloseMenus();
	
	BuffyFrame:StartMoving();
	BuffyFrame:SetUserPlaced(false);
	BuffyFrame.IsMoving = true;
end

function BuffyFrame_OnMouseUp()
	if(BuffyFrame.IsMoving) then
		BuffyFrame:StopMovingOrSizing();
		BuffyFrame.IsMoving = false;
		BuffyFrame_SavePosition();
	end
end

function BuffyFrame_SavePosition()
	local point, _, relativePoint, x, y = BuffyFrame:GetPoint();
	
	A.db.global.Position = {
		x = x,
		y = y,
		Point = point,
		RelativePoint = relativePoint,
		DefaultPosition = false,
	};
end

function BuffyFrame_RestorePosition()
	BuffyFrame:ClearAllPoints();
	
	if(not A.db.global or A.db.global.Position.DefaultPosition) then
		BuffyFrame:SetPoint("LEFT", UIParent, "CENTER", -22, GetScreenHeight() * 0.18);
	else
		local p = A.db.global.Position;
		BuffyFrame:SetPoint(p.Point, UIParent, p.RelativePoint, p.x, p.y);
	end
end

function BuffyFrame_ResetPosition()
	BuffyFrame:ClearAllPoints();
	BuffyFrame:SetPoint("LEFT", UIParent, "CENTER", -22, GetScreenHeight() * 0.18);
	
	local point, _, relativePoint, x, y = BuffyFrame:GetPoint();
	
	A.db.global.Position = {
		x = x,
		y = y,
		Point = point,
		RelativePoint = relativePoint,
		DefaultPosition = true,
	};
	
	BuffyFrame_RestorePosition();
end

function A:ResetPosition()
	BuffyFrame_ResetPosition();
end

local MESSAGE_PATTERN = "|cff3ebfeaBuffy|r %s";
function A:AddMessage(pattern, ...)
	DEFAULT_CHAT_FRAME:AddMessage(MESSAGE_PATTERN:format(string.format(pattern, ...)));
end

function A:GetBinding()
	-- return GetBindingKey("CLICK BuffySpellButtonFrame:LeftButton");
	return A.db.global.Keybind;
end

function A:SetTempBind()
	local key = A:GetBinding();
	if(key) then
		SetOverrideBindingClick(BuffyFrame, true, key, "BuffySpellButtonFrame", "LeftButton");
	end
end

function A:ClearTempBind()
	if(not InCombatLockdown()) then
		ClearOverrideBindings(BuffyFrame);
	end
end

function BuffyKeybindingFrame_OnShow()
	BuffyKeybindingFrame.bindkey = A:GetBinding();
	
	if(not BuffyKeybindingFrame.bindkey) then
		BuffyKeybindingFrame.currentBinding:SetText(BUFFY_CHOOSE_BINDING_TEXT);
	else
		BuffyKeybindingFrame.currentBinding:SetText(BuffyKeybindingFrame.bindkey);
	end
	
	-- BuffyKeybindingFrame.bindingUsedBy:SetText(BUFFY_PRESS_BUTTON_TEXT);
end

function BuffyKeybindingFrame_OnKeydown(self, key)
	if(key == "ESCAPE") then
		BuffyKeybindingFrameOuter:Hide();
	end
	
	if(key == "LSHIFT" or key == "RSHIFT" or key == "LCTRL" or key == "RCTRL" or key == "LALT" or key == "RALT") then
		return;
	end
	
	if(IsShiftKeyDown()) then key = "SHIFT-" .. key; end
	if(IsControlKeyDown()) then key = "CTRL-" .. key; end
	if(IsAltKeyDown()) then key = "ALT-" .. key; end
	
	BuffyKeybindingFrame.bindkey = key;
	BuffyKeybindingFrame.currentBinding:SetText(key);
	
	-- local bindAction = GetBindingAction(key);
	-- local bindActionText = _G["BINDING_NAME_" .. bindAction];
	
	-- if(bindAction ~= "") then
	-- 	BuffyKeybindingFrame.bindingUsedBy:SetFormattedText("%s|n%s", bindAction, bindActionText or "");
	-- else
	-- 	BuffyKeybindingFrame.bindingUsedBy:SetText("Binding not in use");
	-- end
end

function BuffyKeybindingFrame_OnMouseWheel(self, delta)
	if(delta >= 1) then
		BuffyKeybindingFrame_OnKeydown(self, "MOUSEWHEELUP");
	elseif(delta <= -1) then
		BuffyKeybindingFrame_OnKeydown(self, "MOUSEWHEELDOWN");
	end
end

function BuffyKeybindingFrame_OnMouseDown(self, button)
	if(button == "LeftButton") then
		BuffyKeybindingFrame_OnKeydown(self, "BUTTON1");
	elseif(button == "RightButton") then
		BuffyKeybindingFrame_OnKeydown(self, "BUTTON2");
	elseif(button == "MiddleButton") then
		BuffyKeybindingFrame_OnKeydown(self, "BUTTON3");
	else
		BuffyKeybindingFrame_OnKeydown(self, button);
	end
end

function BuffyKeybindingFrameAcceptButton_OnClick()
	if(A:GetBinding()) then A:ClearTempBind(); end
	
	A.db.global.Keybind = BuffyKeybindingFrame.bindkey;
	A:SetTempBind();
	
	BuffyKeybindingFrameOuter:Hide();
end

function BuffySpellButtonFrame_OnClick()
	if(InCombatLockdown()) then
		A:AddMessage("Cannot cast recommended buff while in combat");
	end
end

function BuffyFrame_OnClick(self, button)
	if(button == "RightButton") then
		A:OpenContextMenu(self);
	end
end

function BuffyFrame_OnFadeInPlay(self)
	BuffyFrame:Show();
end

function BuffyFrame_OnFadeOutFinished(self)
	BuffyFrame:Hide();
end