------------------------------------------------------------
-- Buffy by Sonaza
------------------------------------------------------------

local ADDON_NAME, SHARED = ...;

local LibStub = LibStub;
local A = SHARED[1];

local LDB = LibStub:GetLibrary("LibDataBroker-1.1");
local AceDB = LibStub("AceDB-3.0")

local ICON_PATTERN_14 = "|T%s:14:14:0:0|t";
local BUFFY_ICON = "Interface\\Icons\\spell_arcane_invocation";

local TEX_BUFFY_ICON = ICON_PATTERN_14:format(BUFFY_ICON);

function A:InitializeDatabase()
	local defaults = {
		global = {
			Enabled = true,
			ShowInCombat = false,
			ShowWhileMounted = false,
			ShowBuffStatus = true,
			DisableOutside = false,
			
			Keybind = nil,
			
			Position = {
				DefaultPosition = true,
				x = nil,
				y = nil,
				Point = nil,
				RelativePoint = nil,
			},
			
			ExpirationAlertThreshold = 300,
			NoExpirationAlertInCombat = true,
			
			ConsumablesRemind = {
				Enabled = true,
				Flasks = true,
				Runes = false,
				
				SkipStamina = false,
				EnableInParty = false,
				KeybindEnabled = false,
				NotInCombat = true,
			},
			
			Class = {
				Rogue = {
					EnableLethal = true,
					EnableNonlethal = true,
					SkipCrippling = false,
					WoundPoisonPriority = false,
				},
				Shaman = {
					RemindShield = true,	
				},
				Hunter = {
					EnableLoneWolf = true,	
				},
				Priest = {
					RemindShadowform = true,	
				},
				Paladin = {
					RemindRighteousFury = true,	
				},
			},
		},
	};
	
	self.db = AceDB:New(ADDON_NAME .. "DB", defaults);
end

local classOptions = {
	["ROGUE"]	= {
		text = "Rogue Options",
		hasArrow = true,
		notCheckable = true,
		menuList = {
			{
				text = "Rogue Options", isTitle = true, notCheckable = true,
			},
			{
				text = "Enable Alert for Lethal Poisons",
				func = function() A.db.global.Class.Rogue.EnableLethal = not A.db.global.Class.Rogue.EnableLethal; A:UpdateBuffs(); end,
				checked = function() return A.db.global.Class.Rogue.EnableLethal; end,
				isNotRadio = true,
				hasArrow = true,
				menuList = {
					{
						text = "Prioritize Wound Poison over Others",
						func = function() A.db.global.Class.Rogue.WoundPoisonPriority = not A.db.global.Class.Rogue.WoundPoisonPriority; A:UpdateBuffs(); end,
						checked = function() return A.db.global.Class.Rogue.WoundPoisonPriority; end,
						isNotRadio = true,
					},
				},
			},
			{
				text = "Enable Alert for Non-Lethal Poisons",
				func = function() A.db.global.Class.Rogue.EnableNonlethal = not A.db.global.Class.Rogue.EnableNonlethal; A:UpdateBuffs(); end,
				checked = function() return A.db.global.Class.Rogue.EnableNonlethal; end,
				isNotRadio = true,
				hasArrow = true,
				menuList = {
					{
						text = "Skip Crippling Poison",
						func = function() A.db.global.Class.Rogue.SkipCrippling = not A.db.global.Class.Rogue.SkipCrippling; A:UpdateBuffs(); end,
						checked = function() return A.db.global.Class.Rogue.SkipCrippling; end,
						isNotRadio = true,
					},
				},
			},
		},
	},
	["SHAMAN"]	= {
		text = "Shaman Options",
		hasArrow = true,
		notCheckable = true,
		menuList = {
			{
				text = "Shaman Options", isTitle = true, notCheckable = true,
			},
			{
				text = "Enable Alert for Shaman Shield Buff",
				func = function() A.db.global.Class.Shaman.RemindShield = not A.db.global.Class.Shaman.RemindShield; A:UpdateBuffs(); end,
				checked = function() return A.db.global.Class.Shaman.RemindShield; end,
				isNotRadio = true,
			},
		},
	},
	["PRIEST"]	= {
		text = "Priest Options",
		hasArrow = true,
		notCheckable = true,
		menuList = {
			{
				text = "Priest Options", isTitle = true, notCheckable = true,
			},
			{
				text = "Enable Alert for Shadowform",
				func = function() A.db.global.Class.Priest.RemindShadowform = not A.db.global.Class.Priest.RemindShadowform; A:UpdateBuffs(); end,
				checked = function() return A.db.global.Class.Priest.RemindShadowform; end,
				isNotRadio = true,
			},
		},
	},
	["PALADIN"]	= {
		text = "Paladin Options",
		hasArrow = true,
		notCheckable = true,
		menuList = {
			{
				text = "Paladin Options", isTitle = true, notCheckable = true,
			},
			{
				text = "Enable Alert for Righteous Fury",
				func = function() A.db.global.Class.Paladin.RemindRighteousFury = not A.db.global.Class.Paladin.RemindRighteousFury; A:UpdateBuffs(); end,
				checked = function() return A.db.global.Class.Paladin.RemindRighteousFury; end,
				isNotRadio = true,
			},
		},
	},
	["HUNTER"]	= {
		text = "Hunter Options",
		hasArrow = true,
		notCheckable = true,
		menuList = {
			{
				text = "Hunter Options", isTitle = true, notCheckable = true,
			},
			{
				text = "Enable Alerts for Lone Wolf Buffs",
				func = function() A.db.global.Class.Hunter.EnableLoneWolf = not A.db.global.Class.Hunter.EnableLoneWolf; A:UpdateBuffs(); end,
				checked = function() return A.db.global.Class.Hunter.EnableLoneWolf; end,
				isNotRadio = true,
			},
		},
	},
};

StaticPopupDialogs["BUFFY_CUSTOM_EXPIRE_TIME"] = {
	text = "Enter a custom expiration alert threshold in seconds (e.g. 1800 is half an hour):",
	button1 = ACCEPT,
	button2 = CANCEL,
	hasEditBox = 1,
	maxLetters = 4,
	OnAccept = function(self)
		local timeValue = tonumber(self.editBox:GetText());
		if(timeValue ~= nil) then
			A.db.global.ExpirationAlertThreshold = timeValue;
			A:UpdateBuffs();
		end
	end,
	EditBoxOnEnterPressed = function(self)
		local parent = self:GetParent();
		local timeValue = tonumber(parent.editBox:GetText());
		if(timeValue ~= nil) then
			A.db.global.ExpirationAlertThreshold = timeValue;
			A:UpdateBuffs();
		end
		parent:Hide();
	end,
	OnShow = function(self)
		self.editBox:SetFocus();
	end,
	OnHide = function(self)
		ChatEdit_FocusActiveWindow();
		self.editBox:SetText("");
	end,
	timeout = 0,
	exclusive = 0,
	hideOnEscape = 1
};

function A:IsSomeDefaultExpirationTime(t)
	return t == 0 or t == 60 or t == 120 or t == 300 or t == 600 or t == 900; 
end

function A:GetDatabrokerMenuData()
	local legibleClass, class = UnitClass("player");
	
	local locktext = "Unlock";
	
	if(not A.IsFrameLocked) then
		locktext = "Lock";
	end
	
	local keybind, keybindingText = A:GetBinding();
	if(keybind) then
		keybindingText = string.format("Change the keybinding (|cfff5ce47Current %s|r)", keybind);
	else
		keybindingText = "Set a keybind (|cfff15050No binding!|r)";
	end
	
	local expirationAlertTimeText = "Disabled";
	if(self.db.global.ExpirationAlertThreshold > 0) then
		expirationAlertTimeText = A:FormatTime(self.db.global.ExpirationAlertThreshold, true);
	end
	
	local customTimeText = "Custom";
	if(not A:IsSomeDefaultExpirationTime(self.db.global.ExpirationAlertThreshold)) then
		customTimeText = string.format("Custom: %s", A:FormatTime(self.db.global.ExpirationAlertThreshold, true));
	end
	
	local data = {
		{
			text = "Buffy Options", isTitle = true, notCheckable = true,
		},
		{
			text = "Buffy Notifications Enabled",
			func = function() self.db.global.Enabled = not self.db.global.Enabled; A:UpdateBuffs(); end,
			checked = function() return self.db.global.Enabled; end,
			isNotRadio = true,
		},
		{
			text = string.format("%s Buffy Frame", locktext),
			func = function() self.IsFrameLocked = not self.IsFrameLocked; end,
			notCheckable = true,
		},
		{
			text = " ", isTitle = true, notCheckable = true,
		},
		{
			text = keybindingText,
			func = function() BuffyKeybindingFrameOuter:Show(); end,
			notCheckable = true,
		},
		{
			text = " ", isTitle = true, notCheckable = true,
		},
		{
			text = "Notifications", isTitle = true, notCheckable = true,
		},
		{
			text = "Disable Notifications While not in an Instance",
			func = function() self.db.global.DisableOutside = not self.db.global.DisableOutside; A:UpdateBuffs(); end,
			checked = function() return self.db.global.DisableOutside; end,
			isNotRadio = true,
		},
		{
			text = "Show Notifications in Combat",
			func = function() self.db.global.ShowInCombat = not self.db.global.ShowInCombat; A:UpdateBuffs(); end,
			checked = function() return self.db.global.ShowInCombat; end,
			isNotRadio = true,
		},
		{
			text = "Show Notifications While Flying or in a Vehicle",
			func = function() self.db.global.ShowWhileMounted = not self.db.global.ShowWhileMounted; A:UpdateBuffs(); end,
			checked = function() return self.db.global.ShowWhileMounted; end,
			isNotRadio = true,
		},
		{
			text = string.format("Remind About Raid Consumables in Raids", expirationAlertTimeText),
			func = function() self.db.global.ConsumablesRemind.Enabled = not self.db.global.ConsumablesRemind.Enabled; A:UpdateBuffs(); end,
			checked = function() return self.db.global.ConsumablesRemind.Enabled; end,
			isNotRadio = true,
			hasArrow = true,
			menuList = {
				{
					text = "Consumables Remind", isTitle = true, notCheckable = true,
				},
				{
					text = "Enable for Flasks",
					func = function() self.db.global.ConsumablesRemind.Flasks = not self.db.global.ConsumablesRemind.Flasks; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ConsumablesRemind.Flasks; end,
					isNotRadio = true,
				},
				{
					text = "Enable for Augment Runes",
					func = function() self.db.global.ConsumablesRemind.Runes = not self.db.global.ConsumablesRemind.Runes; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ConsumablesRemind.Runes; end,
					isNotRadio = true,
				},
				{
					text = " ", isTitle = true, notCheckable = true,
				},
				{
					text = "Other Options", isTitle = true, notCheckable = true,
				},
				{
					text = "Prioritize Main Stat over Stamina",
					func = function() self.db.global.ConsumablesRemind.SkipStamina = not self.db.global.ConsumablesRemind.SkipStamina; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ConsumablesRemind.SkipStamina; end,
					isNotRadio = true,
				},
				{
					text = "Enable in 5-Man Dungeons",
					func = function() self.db.global.ConsumablesRemind.EnableInParty = not self.db.global.ConsumablesRemind.EnableInParty; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ConsumablesRemind.EnableInParty; end,
					isNotRadio = true,
				},
				{
					text = "Also Enable Keybind for Consumables",
					func = function() self.db.global.ConsumablesRemind.KeybindEnabled = not self.db.global.ConsumablesRemind.KeybindEnabled; A:ClearTempBind(); A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ConsumablesRemind.KeybindEnabled; end,
					isNotRadio = true,
				},
				{
					text = "Do not Alert While in Combat",
					func = function() self.db.global.ConsumablesRemind.NotInCombat = not self.db.global.ConsumablesRemind.NotInCombat; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ConsumablesRemind.NotInCombat; end,
					isNotRadio = true,
				},
			},
		},
		{
			text = string.format("Near Expiration Alert (%s)", expirationAlertTimeText),
			menuList = {
				{
					text = "Near Expiration Alert Options", isTitle = true, notCheckable = true,
				},
				{
					text = "Do not Alert While in Combat",
					func = function() self.db.global.NoExpirationAlertInCombat = not self.db.global.NoExpirationAlertInCombat; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.NoExpirationAlertInCombat; end,
					isNotRadio = true,
				},
				{
					text = " ", isTitle = true, notCheckable = true,
				},
				{
					text = "Select Threshold for Time Remaining", isTitle = true, notCheckable = true,
				},
				{
					text = "Disabled",
					func = function() self.db.global.ExpirationAlertThreshold = 0; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ExpirationAlertThreshold == 0; end,
				},
				{
					text = "1 minute",
					func = function() self.db.global.ExpirationAlertThreshold = 60; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ExpirationAlertThreshold == 60; end,
				},
				{
					text = "2 minutes",
					func = function() self.db.global.ExpirationAlertThreshold = 120; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ExpirationAlertThreshold == 120; end,
				},
				{
					text = "5 minutes",
					func = function() self.db.global.ExpirationAlertThreshold = 300; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ExpirationAlertThreshold == 300; end,
				},
				{
					text = "10 minutes",
					func = function() self.db.global.ExpirationAlertThreshold = 600; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ExpirationAlertThreshold == 600; end,
				},
				{
					text = "15 minutes",
					func = function() self.db.global.ExpirationAlertThreshold = 900; A:UpdateBuffs(); CloseMenus(); end,
					checked = function() return self.db.global.ExpirationAlertThreshold == 900; end,
				},
				{
					text = customTimeText,
					func = function() StaticPopup_Show("BUFFY_CUSTOM_EXPIRE_TIME"); CloseMenus(); end,
					checked = function()
						return not A:IsSomeDefaultExpirationTime(self.db.global.ExpirationAlertThreshold);
					end,
				},
			},
			hasArrow = true,
			notCheckable = true,
		},
		{
			text = " ", isTitle = true, notCheckable = true,
		},
		{
			text = "Other Options", isTitle = true, notCheckable = true,
		},
		{
			text = "Show Raid Buff Status on Databroker",
			func = function() self.db.global.ShowBuffStatus = not self.db.global.ShowBuffStatus; A:UpdateBuffs(); end,
			checked = function() return self.db.global.ShowBuffStatus; end,
			isNotRadio = true,
		},
	};
	
	if(not A.db.global.Position.DefaultPosition) then
		tinsert(data, 4, {
			text = "Reset Position",
			func = function() A:ResetPosition(); end,
			notCheckable = true,
		});
	end
	
	if(classOptions[class] ~= nil) then
		tinsert(data, classOptions[class]);
	else
		tinsert(data, {
			text = "No options for " .. legibleClass,
			notCheckable = true,
			disabled = true,
		});
	end
	
	tinsert(data, {
		text = " ", isTitle = true, notCheckable = true,
	});
	
	tinsert(data, {
		text = "Close Menu",
		func = function() CloseMenus() end,
		notCheckable = true,
	});
	
	return data;
end

function A:OpenContextMenu(parentframe)
	if(not A.ContextMenu) then
		A.ContextMenu = CreateFrame("Frame", ADDON_NAME .. "ContextMenuFrame", parentframe, "UIDropDownMenuTemplate");
	end
	
	A.ContextMenu:SetPoint("BOTTOM", parentframe, "CENTER", 0, 5);
	EasyMenu(A:GetDatabrokerMenuData(), A.ContextMenu, "cursor", 0, 0, "MENU", 5);
end

function A:UpdateDatabrokerText()
	if(self.db.global.ShowBuffStatus) then
		local buffs = A:ScanCurrentBuffs();
		local foundBuffs = 0;
		local numBuffTypes = #A.BUFF_TYPES;
		
		for category, data in pairs(buffs) do
			if(data ~= false) then
				foundBuffs = foundBuffs + 1;
			end
		end
		
		A.databroker.text = string.format("%d/%d |cffffcc00buffs|r", foundBuffs, numBuffTypes);
	else
		A.databroker.text = "Buffy";
	end
end

local function GetClassColor(class)
	return (CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS)[class];
end

-- function A:UnitIsPet(unit)
-- 	return UnitPlayerControlled(unit) and not UnitIsPlayer(unit) and
-- 		  (UnitPlayerOrPetInParty(unit) or UnitPlayerOrPetInRaid(unit) or UnitIsUnit(unit, "pet"));
-- end



function A:InitializeDatabroker()
	A.databroker = LDB:NewDataObject(ADDON_NAME, {
		type = "data source",
		label = "Buffy",
		text = "Buffy",
		icon = BUFFY_ICON,
		OnClick = function(frame, button)
			if(button == "RightButton") then
				GameTooltip:Hide();
				A:OpenContextMenu(frame);
			end
		end,
		OnTooltipShow = function(tooltip)
			if not tooltip or not tooltip.AddLine then return end
			tooltip:AddLine(TEX_BUFFY_ICON .. " |cffffffffBuffy|r");
			tooltip:AddLine(" ");
			
			if(not self.db.global.Enabled) then
				tooltip:AddLine("|cffe05754Notifications Currently Disabled|r");
				tooltip:AddLine(" ");
			end
			
			tooltip:AddLine("|cff78dbfcCurrent Buffs|r");
			
			local buffs = A:ScanCurrentBuffs();
			
			for index, category in ipairs(A.BUFF_TYPES) do
				local data = buffs[category];
				local categoryName = A.BUFF_TYPE_NAMES[category];
				
				if(data == false) then
					tooltip:AddDoubleLine(categoryName, "|cffffffffMissing|r");
				else
					local casterName = data.casterName;
					local classColor = "ffffffff";
					
					local isPet, masterName, petName = A:UnitIsPet(casterName);
					if(not isPet) then
						classColor = GetClassColor(data.casterClass).colorStr;
					else
						casterName = string.format("%s|r's pet", masterName);
						
						local _, class = UnitClass(masterName);
						classColor = GetClassColor(class).colorStr;
					end
					
					local buffname, _, bufficon = GetSpellInfo(data.buff);
					
					tooltip:AddDoubleLine(
						categoryName,
						string.format("%s %s (|c%s%s|r)", ICON_PATTERN_14:format(bufficon), buffname, classColor, casterName),
						nil, nil, nil, 1, 1, 1
					);
				end
			end
			
			tooltip:AddLine(" ");
			tooltip:AddLine("Right-Click |cffffffffShow options|h");
			
			-- Addon.tooltip_open = true;
		end,
		OnLeave = function(frame)
			-- Addon.tooltip_open = false;
		end,
	});

	A:UpdateDatabrokerText();
end